@@ -192,7 +192,7 @@ class KinovaGen3Arm(robot_arms.RobotArm):
         self.connection.execute_existing_action('Home', Base_pb2.REACH_JOINT_ANGLES, wait=wait)
 
     def set_tool(self, offset: Location, mass_kg: float):
-        tool_transform = ControlConfig_pb2.CartesianTransform
+        tool_transform = ControlConfig_pb2.CartesianTransform()
         tool_transform.x = offset.x
         tool_transform.y = offset.y
         tool_transform.z = offset.z
@@ -209,6 +209,62 @@ class KinovaGen3Arm(robot_arms.RobotArm):
     def wait(self, timeout: Optional[float] = None):
         self.connection.wait_for_action_end(timeout)
 
+    def build_joint_angles_from_joints(self, joint_angles: Base_pb2.JointAngles, joints: List[float]):
+        if len(joints) != self.joint_count:
+            raise robot_arms.RobotArmInvalidJointsError(f'Invalid number of joint angles ({len(joints)}), must be {self.joint_count}')
+
+        for joint_id in range(self.joint_count):
+            joint_angle = joint_angles.joint_angles.add()
+            joint_angle.joint_identifier = joint_id
+            joint_angle.value = joints[joint_id]
+
+    def build_joints_from_joint_angles(self, joint_angles: Base_pb2.JointAngles) -> List[float]:
+        return [joint.value for joint in joint_angles.joint_angles]
+
+    def build_pose_from_location(self, pose: Base_pb2.Pose, location: Location):
+        converted_location = location.convert_mm_to_m() if self.position_units == Units.MILLIMETERS else location
+        pose.x = converted_location.x
+        pose.y = converted_location.y
+        pose.z = converted_location.z
+        pose.theta_x = converted_location.rx
+        pose.theta_y = converted_location.ry
+        pose.theta_z = converted_location.rz
+
+    def build_location_from_pose(self, pose: Base_pb2.Pose) -> Location:
+        location = Location(
+            x=pose.x,
+            y=pose.y,
+            z=pose.z,
+            rx=pose.theta_x,
+            ry=pose.theta_y,
+            rz=pose.theta_z,
+        )
+
+        if self.position_units == Units.MILLIMETERS:
+            return location.convert_m_to_mm()
+
+        return location
+
+    def inverse_kinematics(self, location: Location, guess: Optional[List[float]] = None) -> List[float]:
+        joint_angles_guess = self.joint_positions if guess is None else guess
+        ik_data = Base_pb2.IKData()
+        self.build_pose_from_location(ik_data.cartesian_pose, location)
+        self.build_joint_angles_from_joints(ik_data.guess, joint_angles_guess)
+
+        joint_angles = self.connection.client.ComputeInverseKinematics(ik_data)
+
+        return self.build_joints_from_joint_angles(joint_angles)
+
+    def forward_kinematics(self, joints: List[float]) -> Location:
+        if len(joints) != self.joint_count:
+            raise robot_arms.RobotArmInvalidJointsError(f'Invalid number of joint angles ({len(joints)}), must be {self.joint_count}')
+
+        joint_angles = Base_pb2.JointAngles()
+        self.build_joint_angles_from_joints(joint_angles, joints)
+        pose = self.connection.client.ComputeForwardKinematics(joint_angles)
+
+        return self.build_location_from_pose(pose)
+
     def move_to_location(self, location: Location,
                          velocity: Optional[float] = None, acceleration: Optional[float] = None, frame: Optional[Frame] = None,
                          relative: bool = False, wait: bool = True, timeout: Optional[float] = None):
